""" Main module for spikely application

This module is home to the entry point for the application, launch_spikely(),
called when the user invokes spikely from the command line. In response,
spikely_main instantiates the PyQt Application class (QApplication) and the
associated widget hierarchy starting w/ SpikelyMainWindow at the top.  Once
these tasks are performed, execution shifts to the xxx_view.py and xxx_menu.py
modules whose methods are called in response to user interactions with the UI.

"""
import sys

import pkg_resources
from PyQt5 import QtWidgets, QtCore, QtGui

from spikely import (
    PipelineModel, PipelineView, ParameterView, ParameterModel,
    OperationView, file_menu, help_menu, tool_bar, __version__)


class SpikelyMainWindow(QtWidgets.QMainWindow):
    # Parent UI for application delegates to subwindow views/models
    def __init__(self):
        super().__init__()

        # Subwindows Views need underlying Model references
        self._parameter_model = ParameterModel()
        self._pipeline_model = PipelineModel(self._parameter_model)
        self._init_ui()

    def _init_ui(self):
        self.setWindowTitle("spikely")
        self.setGeometry(100, 100, 1280, 512)

        # Disable maximize button on title bar
        # self.setWindowFlags(self.windowFlags()
        #                     & ~QtCore.Qt.WindowMaximizeButtonHint)

        try:
            spikely_png_path = pkg_resources.resource_filename(
                "spikely.resources", "spikely.png")
        except KeyError:
            print(
                "<<spikely error: Failed to find spikely.png in resource "
                "directory>>", file=sys.stderr,)
        else:
            self.setWindowIcon(QtGui.QIcon(spikely_png_path))

        self.statusBar().addPermanentWidget(
            QtWidgets.QLabel("Version " + __version__))

        menu_bar = self.menuBar()
        menu_bar.addMenu(file_menu.create_file_menu(
            self, self._pipeline_model))
        menu_bar.addMenu(help_menu.create_help_menu(self))

        bar = tool_bar.create_tool_bar(self)
        self.addToolBar(QtCore.Qt.RightToolBarArea, bar)

        main_frame = QtWidgets.QFrame()
        self.setCentralWidget(main_frame)
        main_frame.setLayout(QtWidgets.QVBoxLayout())

        pipe_param_splitter = QtWidgets.QSplitter()
        pipe_param_splitter.setChildrenCollapsible(False)

        # Subwindows for element pipeline and selected element parameters
        pipe_param_splitter.addWidget(
            PipelineView(self._pipeline_model, self._parameter_model))
        pipe_param_splitter.addWidget(
            ParameterView(self._pipeline_model, self._parameter_model))
        pipe_param_splitter.setSizes([256, 1024])
        main_frame.layout().addWidget(pipe_param_splitter)

        # Subwindow at bottom for pipeline operations (run, clear, queue)
        main_frame.layout().addWidget(
            OperationView(self._pipeline_model, self._parameter_model))

        # Stretches pipeline and parameter widgets down as window grows
        main_frame.layout().setStretch(0, 1)


def launch_spikely():
    app = QtWidgets.QApplication(sys.argv)
    win = SpikelyMainWindow()
    win.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    launch_spikely()



# Constants and helper functions used by other spikely modules
import importlib
import sys

from PyQt5 import QtWidgets

from spikely import SpikeElement

# Duration in milliseconds of timeout for temporary status messages
STATUS_MSG_TIMEOUT = 3500

# Identifier to get elem object from pipeline model data()
ELEMENT_ROLE = 0x100

# Column IDs used by QTableView to display elem parameter data
PARAM_COL, TYPE_COL, VALUE_COL = 0, 1, 2


def get_main_window() -> QtWidgets.QMainWindow:
    """Returns the app's main window for use as message box parent."""
    for widget in QtWidgets.QApplication.instance().topLevelWidgets():
        if isinstance(widget, QtWidgets.QMainWindow):
            return widget

    # It is a dark day if we end up here
    print("<<spikely fatal error: Failed to find QMainWindow.>>",
          file=sys.stderr)
    sys.exit()


def cvt_elem_to_dict(elem: SpikeElement) -> dict:
    """Converts element to dictionary to enable JSON encoding.

    Elements cannot be directly json encoded, so this function stores and
    element's class and module names along its instance data in a dictionary.
    The dictionary can be encoded, and when decoded a new instance of the
    element can be instantiated.

    json encoded elements are saved to files to allow pipeline saves and loads,
    and also used to transfer pipelines as strings between processes.

    """

    if not isinstance(elem, SpikeElement):
        raise TypeError("elem must be a SpikeElement object")

    elem_dict = {
        "element_cls_name": elem.__class__.__name__,
        "element_mod_name": elem.__module__,
        "spif_cls_name": elem.spif_class.__name__,
        "spif_mod_name": elem.spif_class.__module__,
        "param_list": elem.param_list,
    }

    return elem_dict


def cvt_dict_to_elem(elem_dict: dict) -> SpikeElement:
    """ Converts an element dictionary into an element.

    Used as part of the json encode/decode process, this method "reconstitutes"
    an element from a dictionary of element instance data that had been json
    encoded.

    """

    if not isinstance(elem_dict, dict):
        raise TypeError("elem_dict must be a dict object")

    elem_mod = importlib.import_module(elem_dict["element_mod_name"])
    elem_cls = getattr(elem_mod, elem_dict["element_cls_name"])
    spif_mod = importlib.import_module(elem_dict["spif_mod_name"])
    spif_cls = getattr(spif_mod, elem_dict["spif_cls_name"])

    if not spif_cls.installed:
        # Abort if spif_class is no longer installed on system
        raise ValueError(
            f"Cannot create {elem_dict['spif_cls_name']} - "
            f" not installed on users's system"
        )

    elem = elem_cls(spif_cls)

    elem_param_name_set = {param["name"] for param in elem.param_list}

    dict_param_name_set = {param["name"] for param in elem_dict["param_list"]}

    if not dict_param_name_set.issubset(elem_param_name_set):
        # Abort if the old param list is not a subset of new one
        raise ValueError(
            f"Cannot create {elem_dict['spif_cls_name']} - "
            f" saved version incompatible with current version"
        )

    elem.param_list = elem_dict["param_list"]

    return elem



import json
import os
import sys

from spikely import config as cfg


def run(elem_list_str):

    elem_jdict_list = json.loads(elem_list_str)
    elem_list = [cfg.cvt_dict_to_elem(elem_jdict)
                 for elem_jdict in elem_jdict_list]

    payload = None
    last_index = len(elem_list) - 1
    for index, elem in enumerate(elem_list):
        if index == last_index:
            next_elem = None
        else:
            next_elem = elem_list[index + 1]
        payload = elem.run(payload, next_elem)


def main():
    run(sys.argv[1])

    # Turns out that this is a very important call
    os._exit(1)


if __name__ == '__main__':
    main()



# imports ensure SpikeElement.__subclasses__() works properly
# from . import spike_element     # noqa: F401
# from . import recording_extractor         # noqa: F401
# from . import sorter            # noqa: F401
# from . import preprocessor      # noqa: F401



# Implements MVC view associated with pipeline operations (run,clear, queue)

from PyQt5 import QtWidgets

from spikely import config


class OperationView(QtWidgets.QGroupBox):

    def __init__(self, pipeline_model, parameter_model):
        super().__init__("Command Pipeline")  # Group box label
        self._pipeline_model = pipeline_model
        self._parameter_model = parameter_model

        self._init_ui()

    def _init_ui(self):

        # Disable/Enable command buttons on empty/non-empty pipeline state
        self._pipeline_model.rowsInserted.connect(self._pipeline_changed)
        self._pipeline_model.rowsRemoved.connect(self._pipeline_changed)
        self._pipeline_model.modelReset.connect(self._pipeline_changed)

        self.setLayout(QtWidgets.QHBoxLayout())

        self._run_btn = QtWidgets.QPushButton("Run")
        self._run_btn.setStatusTip('Command the pipeline to run - executes '
            'from top to bottom')  # noqa: E128
        self._run_btn.setEnabled(False)
        self._run_btn.clicked.connect(self._pipeline_model.run)
        self.layout().addWidget(self._run_btn)

        self._clear_btn = QtWidgets.QPushButton("Clear")
        self._clear_btn.setStatusTip('Command the pipeline to clear - '
            'removes all elements')  # noqa: E128
        self._clear_btn.setEnabled(False)
        self._clear_btn.clicked.connect(self._pipeline_model.clear)
        self.layout().addWidget(self._clear_btn)

        self._queue_btn = QtWidgets.QPushButton("Queue")
        self._queue_btn.setStatusTip('Adds pipeline to queue for '
            'batch processing - not implemented')  # noqa: E128
        self._queue_btn.clicked.connect(self._queue_clicked)
        self.layout().addWidget(self._queue_btn)
        self._queue_btn.setEnabled(False)

    def _queue_clicked(self):
        config.get_main_window().statusBar().showMessage(
            "Queue not implemented", config.STATUS_MSG_TIMEOUT)

    def _pipeline_changed(self, parent=None, first=None, last=None):
        enabled = self._pipeline_model.rowCount(None) > 0
        self._run_btn.setEnabled(enabled)
        self._clear_btn.setEnabled(enabled)
        # self._queue_btn.setEnabled(enabled)



from spiketoolkit.postprocessing import export_to_phy
import numpy as np

class PhyExporter:

    exporter_name = 'PhyExporter'
    installed = True
    exporter_gui_params = [
        {'name': 'save_path', 'type': 'folder', 'title': "Save path"},
        {'name': 'n_comp', 'type': 'int', 'value': 3, 'default': 3, 'title': "Number of principal components"},
        {'name': 'electrode_dimensions', 'type': 'int_list', 'value': None, 'default': None, 'title': "If electrode locations are 3D, it indicates the 2D dimensions to use as channel location."},
        {'name': 'grouping_property', 'type': 'str', 'value': None, 'default': None, 'title':
            "Property to group channels. E.g. if the recording extractor has the 'group' property and 'grouping_property' is 'group', then waveforms are computed group-wise."},
        {'name': 'ms_before', 'type': 'float', 'value': 1., 'default': 1., 'title': "Time period in ms to cut waveforms before the spike events."},
        {'name': 'ms_after', 'type': 'float', 'value': 2., 'default': 2., 'title': "Time period in ms to cut waveforms after the spike events."},
        {'name': 'dtype', 'type': 'np.dtype', 'value': None, 'default': None, 'title': "The dtype of underlying data (int16, float32, etc.)"},
        {'name': 'amp_method', 'type': 'str', 'value': 'absolute', 'default': 'absolute', 'title': "If 'absolute' (default), amplitudes are absolute amplitudes in uV are returned. If 'relative', amplitudes are returned as ratios between waveform amplitudes and template amplitudes."},
        {'name': 'amp_peak', 'type': 'str', 'value': 'both', 'default': 'both', 'title': "If maximum channel has to be found among negative peaks ('neg'), positive ('pos') or both ('both' - default)."},
        {'name': 'amp_frames_before', 'type': 'int', 'value': 3, 'default': 3, 'title': "Frames before peak to compute amplitude."},
        {'name': 'amp_frames_after', 'type': 'int', 'value': 3, 'default': 3, 'title': "Frames after peak to compute amplitude."},
        {'name': 'max_spikes_for_pca', 'type': 'int', 'value': 100000, 'default': 100000, 'title': "The maximum number of waveforms used to compute PCA. If 'inf', it will use all the waveforms."},
        {'name': 'write_waveforms', 'type': 'bool', 'value': False, 'default': False, 'title': "If True, waveforms are saved as waveforms.npy."},
        {'name': 'seed', 'type': 'int', 'value': 0, 'default': 0, 'title': "Random seed for extracting waveforms and pcs."},
    ]
    mode = 'folder'

    @staticmethod
    def write_sorting(recording, sorting, save_path, n_comp=3, electrode_dimensions=None,
                      grouping_property=None, ms_before=1., ms_after=2., dtype=None, amp_method='absolute', amp_peak='both',
                      amp_frames_before=3, amp_frames_after=3, max_spikes_for_pca=100000,
                      recompute_info=True, save_features_props=False, write_waveforms=False, verbose=False,
                      seed=0):

        export_to_phy(recording, sorting, save_path, n_comp, electrode_dimensions,
                      grouping_property, ms_before, ms_after, dtype, amp_method, amp_peak,
                      amp_frames_before, amp_frames_after, max_spikes_for_pca,
                      recompute_info, save_features_props, write_waveforms, verbose,
                      seed)



import copy
import inspect
import os

import pkg_resources
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from . import exporterlist as sp_exl
from . import spike_element as sp_spe


class SortingExporter(sp_spe.SpikeElement):
    @staticmethod
    def get_installed_spif_cls_list():
        return sp_exl.exporters_list

    @staticmethod
    def get_display_name_from_spif_class(spif_class):
        return spif_class.exporter_name

    def __init__(self, spif_class):
        super().__init__(spif_class)

        self._display_name = self.get_display_name_from_spif_class(spif_class)

        if QtWidgets.QApplication.instance():
            self._display_icon = QtGui.QIcon(
                pkg_resources.resource_filename(
                    'spikely.resources', 'exporter.png'))
        else:
            self._display_icon = None

        self.param_list = copy.deepcopy(spif_class.exporter_gui_params)

    @property
    def display_name(self):
        return self._display_name

    @property
    def display_icon(self):
        return self._display_icon

    def run(self, payload, next_elem):
        sorting_list = payload[0]
        recording = payload[2]

        if self.display_name == 'NwbSortingExporter':
            nwbfile_kwargs = {}

        for i, sorting in enumerate(sorting_list):
            params_dict = {}
            params_dict['sorting'] = sorting

            if 'recording' in inspect.signature(
                    self.spif_class.write_sorting).parameters:
                params_dict['recording'] = recording
            elif 'sampling_frequency' in inspect.signature(
                    (self.spif_class.write_sorting)).parameters:
                params_dict['sampling_frequency'] = \
                    recording.get_sampling_frequency()

            for param in self.param_list:
                param_name = param['name']
                param_value = param['value']

                if param_name == 'save_path':
                    if(len(sorting_list) == 1):
                        param_value = param_value
                    else:
                        path, file_name = os.path.split(param_value)
                        param_value = path + str(i) + '_' + file_name

                if param_name == 'identifier':
                    nwbfile_kwargs[param_name] = param_value

                if param_name == 'session_description':
                    nwbfile_kwargs[param_name] = param_value
                params_dict[param_name] = param_value

            if self.display_name == 'NwbSortingExporter':
                params_dict['nwbfile_kwargs'] = nwbfile_kwargs

            print("Exporting to " + params_dict['save_path'])
            self.spif_class.write_sorting(**params_dict)



# The view-control widget set for constructing the active pipeline.

from PyQt5 import QtWidgets

from spikely import StdElementPolicy, SpikeElement, config


class PipelineView(QtWidgets.QGroupBox):

    def __init__(self, pipeline_model, parameter_model):
        super().__init__("Construct Pipeline")

        self._pipeline_model = pipeline_model
        self._pipeline_view = QtWidgets.QListView(self)
        self._parameter_model = parameter_model
        self._element_policy = StdElementPolicy()

        self._init_ui()

    def _init_ui(self):
        # Assembles the individual widgets into the widget-set.

        # The PipelineView consists of three separate UI assemblies
        # stacked top to bottom: element selection, pipeline element list,
        # and pipeline element commands (move up, delete, move down)

        # Lay out view from top to bottom of group box
        self.setLayout(QtWidgets.QVBoxLayout())

        self.layout().addWidget(self._element_selection())
        self.layout().addWidget(self._pipeline_list())
        self.layout().addWidget(self._pipeline_commands())

    def _element_selection(self):
        # UI to select for and add elements to pipeline.

        # The UI for element selection combines a combo box for the stages
        # (e.g., RecordingExtractor) and one for the corresponding
        # SpikeInterface classes (e.g., MEArecRecordingExtractor) - the two
        # pieces of information required to instantiate the SpikeElement
        # inserted into the pipeline.

        ui_frame = QtWidgets.QFrame()
        ui_frame.setLayout(QtWidgets.QHBoxLayout())

        # Out of order declaration needed as forward reference
        spif_cbx = QtWidgets.QComboBox(self)
        spif_cbx.setStatusTip('Choose an element to be added to the '
            'pipeline - listed for current element category')  # noqa: E128

        elem_cbx = QtWidgets.QComboBox()
        elem_cbx.setStatusTip('Choose an element category to list the '
            'specific elements available within that category')  # noqa: E128

        ui_frame.layout().addWidget(elem_cbx)

        # Change spif_cbx contents when user makes elem_cbx selection
        def _elem_cbx_changed(index):
            spif_cbx.clear()
            element_cls = elem_cbx.itemData(index)
            # SpikeElement subclasses tasked w/ generating spif class lists
            for spif_cls in element_cls.get_installed_spif_cls_list():
                spif_cbx.addItem(
                    element_cls.get_display_name_from_spif_class(
                        spif_cls), spif_cls)
        elem_cbx.currentIndexChanged.connect(_elem_cbx_changed)

        # Note: cbx instantiation order matters for initial m/v signalling

        # All elements are subclasses of SpikeElement, but element policy
        # determines which ones are available to the user
        elem_classes = [cls for cls in SpikeElement.__subclasses__()
                        if self._element_policy.is_cls_available(cls)]
        # Element (subclass) order is arbitrary, so sort by policy order
        elem_classes.sort(key=lambda e: self._element_policy.cls_order_dict[e])
        # Now that elements are sorted and filtered, set the combo box

        for cls in elem_classes:
            display_name = self._element_policy.get_cls_display_name(cls)
            elem_cbx.addItem(display_name + 's', cls)
        elem_cbx.setCurrentIndex(0)

        ui_frame.layout().addWidget(spif_cbx)

        add_button = QtWidgets.QPushButton("Add Element")
        add_button.setStatusTip('Add selected element to the pipeline - '
            'element will be inserted in category order')  # noqa: E128

        def _add_element_clicked():
            if spif_cbx.currentIndex() > -1:
                # Classes stored as cbx user data enables object creation
                spif_class = spif_cbx.currentData()
                element_class = elem_cbx.currentData()
                element = element_class(spif_class)
                self._pipeline_model.add_element(element)
        add_button.clicked.connect(_add_element_clicked)

        ui_frame.layout().addWidget(add_button)

        return ui_frame

    def _pipeline_list(self):
        # MVC in action - connect View (widget) to Model
        self._pipeline_view.setModel(self._pipeline_model)

        self._pipeline_view.setSelectionMode(
            QtWidgets.QAbstractItemView.SingleSelection)

        # Links element (parameter_view) and pipeline (pipeline_view) views
        def list_selection_changed(selected, deselected):
            if selected.indexes():
                # Retrieve selected element from pipeline model
                element = self._get_selected_element()
                # Link selected element to element property editor
                self._parameter_model.element = element
            else:
                self._parameter_model.element = None
        self._pipeline_view.selectionModel().selectionChanged.connect(
            list_selection_changed)

        return self._pipeline_view

    def _pipeline_commands(self):
        # Operations applied to the pipeline as a whole (Run, Clear, Queue)
        ui_frame = QtWidgets.QFrame()
        ui_frame.setLayout(QtWidgets.QHBoxLayout())

        # Move Up element button and associated action
        mu_btn = QtWidgets.QPushButton("Move Up")
        mu_btn.setStatusTip('Move selected element up one step in the '
            'pipeline - cross element category moves barred')  # noqa: E128
        ui_frame.layout().addWidget(mu_btn)

        def move_up_clicked():
            element = self._get_selected_element()
            if element is None:
                config.get_main_window().statusBar().showMessage(
                    "Nothing to move up", config.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.move_up(element)
        mu_btn.clicked.connect(move_up_clicked)

        # Move Down element button and associated action
        md_btn = QtWidgets.QPushButton("Move Down")
        md_btn.setStatusTip('Move selected element down one step in the '
            'pipeline - cross element category moves barred')  # noqa: E128
        ui_frame.layout().addWidget(md_btn)

        def move_down_clicked():
            element = self._get_selected_element()
            if element is None:
                config.get_main_window().statusBar().showMessage(
                    "Nothing to move down", config.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.move_down(element)
        md_btn.clicked.connect(move_down_clicked)

        # Delete element button and associated action
        de_btn = QtWidgets.QPushButton("Delete")
        de_btn.setStatusTip('Delete the selected element in the pipeline')
        ui_frame.layout().addWidget(de_btn)

        def delete_clicked():
            element = self._get_selected_element()
            if element is None:
                config.get_main_window().statusBar().showMessage(
                    "Nothing to delete", config.STATUS_MSG_TIMEOUT)
            else:
                self._pipeline_model.delete(element)
        de_btn.clicked.connect(delete_clicked)

        return ui_frame

    def _get_selected_element(self):
        # Convenience function to retrieve selected element in pipe view
        element = None
        model = self._pipeline_view.selectionModel()
        if model.hasSelection():
            element = self._pipeline_model.data(
                model.selectedIndexes()[0], config.ELEMENT_ROLE)
        return element



# Application help menu construction and execution

import webbrowser

from PyQt5 import QtWidgets


def create_help_menu(main_window: QtWidgets.QMainWindow) -> QtWidgets.QMenu:
    help_menu = QtWidgets.QMenu('&Help', main_window)
    _create_help_actions(help_menu, main_window)
    return help_menu


def _create_help_actions(menu, win):
    file_actions = [
         ('Documentation', 'Ctrl+D',
          'Open spikely documentation link in web browser',
          _open_doc_browser)
        ]

    for name, shortcut, statustip, signal in file_actions:
        action = QtWidgets.QAction(name, win)
        action.setShortcut(shortcut)
        action.setStatusTip(statustip)
        action.triggered.connect(signal)
        menu.addAction(action)


def _open_doc_browser():
    webbrowser.open('https://spikely.readthedocs.io/en/latest/index.html#a-simple-extracellur-data-processing-application-based-on-spikeinterface)')  # noqa: E501



from abc import ABC, abstractmethod


class ElementPolicy(ABC):

    def __init__(self, required_cls_list, cls_order_dict,
                 cls_display_name_dict):
        self._required_cls_list = required_cls_list
        self._cls_order_dict = cls_order_dict
        self._cls_display_name_dict = cls_display_name_dict

    @abstractmethod
    def is_cls_available(self, cls):
        pass

    @abstractmethod
    def is_cls_singleton(self, cls):
        pass

    @property
    def required_cls_list(self):
        return self._required_cls_list

    @property
    def cls_order_dict(self):
        return self._cls_order_dict

    def get_cls_display_name(self, cls):
        return self._cls_display_name_dict.get(cls, str(cls))



"""Standalone application launched by spikely to manage pipeline execution.

To support parallelism each spikely pipeline is executed in its own process
(QProcess).  pipeman is launched by spikely as a detached process passing in
the pipeline in as a json encoded string in sys.argv[1].  In turn, pipeman
creates a child process of its own, that actually executes the pipeline
(piperun).

Creating a child process of its own allows pipeman to catch and display the
stdout/stderr of piperun, and allows the user to kill piperun without killing
pipeman.

"""
import sys
import locale

import pkg_resources
from PyQt5 import QtCore, QtWidgets, QtGui

from spikely import version, config


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.process = QtCore.QProcess(self)
        self.process.setProcessChannelMode(QtCore.QProcess.MergedChannels)
        self.process.readyReadStandardOutput.connect(self.stdout_ready)

        self._init_ui()

        piperun_path = pkg_resources.resource_filename(
            'spikely.pipeman', 'piperun.py')
        self.process.start('python', ['-u', piperun_path, sys.argv[1]])

        if self.process.state() == QtCore.QProcess.Starting \
                or self.process.state() == QtCore.QProcess.Running:
            self.cancel_btn.setDisabled(False)
            self.process.finished.connect(self._process_finished)

    def _process_finished(self, exit_status):
        self.cancel_btn.setDisabled(True)

    def _init_ui(self):
        self.setWindowTitle("spikely pipeline manager")
        self.resize(640, 384)

        self.statusBar().addPermanentWidget(
            QtWidgets.QLabel("Version " + version.__version__))

        main_frame = QtWidgets.QFrame()
        self.setCentralWidget(main_frame)
        main_frame.setLayout(QtWidgets.QVBoxLayout())

        self.output = QtWidgets.QTextEdit(self)
        self.output.setReadOnly(True)
        self.output.setAcceptRichText(False)
        self.output.setStyleSheet(
            "QTextEdit { color: green; background-color: black; }")
        self.output.setWordWrapMode(QtGui.QTextOption.NoWrap)
        main_frame.layout().addWidget(self.output)

        self.cancel_btn = QtWidgets.QPushButton('Terminate Process')
        btn_box = QtWidgets.QHBoxLayout()
        btn_box.addStretch(1)
        btn_box.addWidget(self.cancel_btn)
        btn_box.addStretch(1)
        self.cancel_btn.setDisabled(True)
        self.cancel_btn.clicked.connect(self.process.kill)
        main_frame.layout().addLayout(btn_box)

    def append(self, text):
        self.output.append(text)

    def stdout_ready(self):
        text = bytearray(self.process.readAllStandardOutput())\
            .decode(locale.getdefaultlocale()[1])

        self.append(text)

    def closeEvent(self, event):
        """Overrides QMainWindow method for confirmation before exiting"""
        if self.process.state() == QtCore.QProcess.Running:
            reply = QtWidgets.QMessageBox.question(
                config.get_main_window(), 'Exiting', 'Exiting will terminate'
                ' pipeline execution.  Are you sure you want to exit?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)

            if reply == QtWidgets.QMessageBox.No:
                event.ignore()


def main():
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow()
    win.show()
    sys.exit(app.exec_())

    # config.get_main_window().statusBar().showMessage(
    #     "Error Message", config.STATUS_MSG_TIMEOUT)


if __name__ == '__main__':
    main()



# Python
import numpy as np
import copy
# PyQt
from PyQt5 import QtGui
import pkg_resources
from PyQt5 import QtWidgets
# spikely
from . import spike_element as sp_spe
import spikeextractors as se


class RecordingExtractor(sp_spe.SpikeElement):
    @staticmethod
    def get_installed_spif_cls_list():
        return se.installed_recording_extractor_list

    @staticmethod
    def get_display_name_from_spif_class(spif_class):
        return spif_class.extractor_name

    def __init__(self, spif_class):
        super().__init__(spif_class)

        self._display_name = self.get_display_name_from_spif_class(spif_class)

        if QtWidgets.QApplication.instance():
            self._display_icon = QtGui.QIcon(
                pkg_resources.resource_filename(
                    'spikely.resources', 'extractor.png'))
        else:
            self._display_icon = None

        self._param_list = copy.deepcopy(spif_class.extractor_gui_params)

        probe_path_dict = {
            'name': 'probe_path', 'type': 'file',
            'value': None, 'default': None,
            'title': 'Path to probe file (.csv or .prb)'
        }
        self._param_list.append(probe_path_dict)

        self._param_list.append({
            'name': 'channel_map', 'type': 'int_list', 'value': None,
            'default': None, 'title': "List of channel ids for underlying \
            channels to be be mapped. If None, then uses default ordering."})

        self._param_list.append({
            'name': 'channel_groups', 'type': 'int_list', 'value': None,
            'default': None, 'title': "List of channel groups of the \
            underlying channels. If None, then no groups given."})

    @property
    def display_name(self):
        return self._display_name

    @property
    def display_icon(self):
        return self._display_icon

    def run(self, payload, next_elem):
        spif_params_dict = {}
        probe_file = None
        for param in self.param_list:
            if param['name'] == 'probe_path':
                probe_file = param['value']
            elif param['name'] == 'channel_map':
                channel_map = param['value']
            elif param['name'] == 'channel_groups':
                channel_groups = param['value']
            else:
                spif_params_dict[param['name']] = param['value']

        recording = self._spif_class(**spif_params_dict)

        if probe_file:
            recording = recording.load_probe_file(
                probe_file, channel_map, channel_groups)
        else:
            if channel_map:
                assert np.all([
                    chan in channel_map for chan in
                    recording.get_channel_ids()]), "all channel_ids in " \
                        "'channel_map' must be in recording channel ids"
                recording = se.SubRecordingExtractor(
                    recording, channel_ids=channel_map)
            if channel_groups:
                recording.set_channel_groups(
                    recording.get_channel_ids(), channel_groups)

        return recording



# Python
import inspect
import shutil
import copy
from pathlib import Path
# PyQt
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import pkg_resources
# spikely
from . import spike_element as sp_spe
import spikeextractors as se
import spiketoolkit as st


class Curator(sp_spe.SpikeElement):
    @staticmethod
    def get_installed_spif_cls_list():
        return st.curation.installed_curation_list

    @staticmethod
    def get_display_name_from_spif_class(spif_class):
        return spif_class.curator_name

    def __init__(self, spif_class):
        super().__init__(spif_class)

        self._display_name = self.get_display_name_from_spif_class(spif_class)

        if QtWidgets.QApplication.instance():
            self._display_icon = QtGui.QIcon(
                pkg_resources.resource_filename(
                    'spikely.resources', 'curator.png'))
        else:
            self._display_icon = None

        self.param_list = copy.deepcopy(spif_class.curator_gui_params)

    @property
    def display_name(self):
        return self._display_name

    @property
    def display_icon(self):
        return self._display_icon

    def run(self, payload, next_element):

        sorting_list = payload[0]
        output_folder_str = payload[1]
        recording = payload[2]

        if not next_element:
            output_folder_str_new = output_folder_str + '_curated'
            output_folder = Path(output_folder_str_new).absolute()
            if output_folder.is_dir():
                shutil.rmtree(output_folder)
            output_folder.mkdir()

        curated_sorting_list = []
        for i, sorting in enumerate(sorting_list):
            params_dict = {}
            params_dict['sorting'] = sorting

            if 'recording' in \
                    inspect.signature(self.spif_class).parameters:
                params_dict['recording'] = recording
            elif 'sampling_frequency' in \
                    inspect.signature(self.spif_class).parameters:
                params_dict['sampling_frequency'] = \
                    recording.get_sampling_frequency()

            for param in self.param_list:
                param_name = param['name']
                param_value = param['value']
                params_dict[param_name] = param_value

            curated_sorting = self.spif_class(**params_dict)
            curated_sorting_list.append(curated_sorting)

            if not next_element:
                print("No Sorting Exporter chosen. Defaulting to "
                      "the .npz format.")
                if len(sorting_list) == 1:
                    file_name = 'curated_output.npz'
                else:
                    file_name = str(i) + '_curated_output.npz'

                se.NpzSortingExtractor.write_sorting(curated_sorting,
                    output_folder / file_name)  # noqa: E128
                print("Saved curated results to " + str(output_folder))

        return curated_sorting_list, output_folder_str, recording



# Python
import copy
import pkg_resources
# PyQt
from PyQt5 import QtGui
from PyQt5 import QtWidgets
# spikely
from . import spike_element as sp_spe
import spiketoolkit as st


class Preprocessor(sp_spe.SpikeElement):
    @staticmethod
    def get_installed_spif_cls_list():
        return st.preprocessing.preprocessinglist. \
            installed_preprocessers_list

    @staticmethod
    def get_display_name_from_spif_class(spif_class):
        return spif_class.preprocessor_name

    def __init__(self, spif_class):
        super().__init__(spif_class)

        self._display_name = self.get_display_name_from_spif_class(spif_class)

        if QtWidgets.QApplication.instance():
            self._display_icon = QtGui.QIcon(
                pkg_resources.resource_filename(
                    'spikely.resources', 'preprocessor.png'))
        else:
            self._display_icon = None

        self._param_list = copy.deepcopy(spif_class.preprocessor_gui_params)

    @property
    def display_name(self):
        return self._display_name

    @property
    def display_icon(self):
        return self._display_icon

    def run(self, payload, next_elem):
        spif_param_dict = {
            param['name']: param['value'] for param in self.param_list}
        spif_param_dict['recording'] = payload
        pp = self._spif_class(**spif_param_dict)
        return pp



import re

import numpy as np
import PyQt5.QtCore as QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from spikely import config


# An MVC model representation of an element's parameters
class ParameterModel(QtCore.QAbstractTableModel):

    def __init__(self):
        self._element = None
        super().__init__()

    # Pythonic setters/getters
    @property
    def element(self):
        return self._element

    @element.setter
    def element(self, element):
        # Ensures dependent views are signaled on element changes
        self.beginResetModel()
        self._element = element
        self.endResetModel()

    #
    # QAbstractTableModel Methods
    #

    # Count of parameters associated with an element instance
    def rowCount(self, parent=QtCore.QModelIndex()):
        return 0 if self._element is None else len(self._element.param_list)

    # Number of display columns: Parameter, Type, Value
    def columnCount(self, parent=QtCore.QModelIndex()):
        return 3

    # Sets UI policy for columns in table view for element parameters
    def flags(self, mod_index):
        column_flags = QtCore.QAbstractTableModel.flags(self, mod_index)
        col = mod_index.column()

        # Parameter and Type column cells are read only
        if col == config.PARAM_COL or col == config.TYPE_COL:
            column_flags ^= QtCore.Qt.ItemIsSelectable
        elif col == config.VALUE_COL:
            column_flags |= QtCore.Qt.ItemIsEditable
        return column_flags

    # Called by Views, get element parameter data based on index and role
    def data(self, mod_index, role=QtCore.Qt.DisplayRole):
        col, row = mod_index.column(), mod_index.row()
        param_dict = self._element.param_list[row]

        # If no actual result return empty value indicator
        result = QtCore.QVariant()

        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            if col == config.PARAM_COL:
                result = param_dict['name']
            elif col == config.TYPE_COL:
                result = param_dict['type']
            elif col == config.VALUE_COL:
                if 'value' in param_dict.keys():
                    result = str(param_dict['value']).strip()
                    if not result and 'default' in param_dict.keys():
                        result = str(param_dict['default'])

        elif role == QtCore.Qt.ToolTipRole:
            if col == config.PARAM_COL and 'title' in param_dict.keys():
                result = param_dict['title']
            elif col == config.TYPE_COL:
                type_str = param_dict['type']
                if type_str == 'int':
                    result = 'integer'
                elif type_str == 'float':
                    result = 'floating point'
                elif type_str == 'bool':
                    result = 'boolean'
                elif type_str == 'int_list':
                    result = 'list of integers'
                elif type_str == 'str':
                    result = 'text string'
                elif type_str == 'file':
                    result = 'pathname of file'
                elif type_str == 'folder':
                    result = 'pathname of folder'
                elif type_str == 'file_or_folder':
                    result = 'pathname of file or folder'
                elif type_str == 'int_list_list':
                    result = 'list of a int_lists'
                elif type_str == 'np.dtype':
                    result = 'numpy dtype object (e.g., int32)'

        # Paints cell red if mandatory parameter value is missing.
        # Mandatory parameters are those with no default keys
        elif role == QtCore.Qt.BackgroundRole:
            if col == config.VALUE_COL:
                if ('value' not in param_dict.keys() and
                        'default' not in param_dict.keys()):
                    result = QtGui.QBrush(QtGui.QColor(255, 192, 192))

        return result

    def headerData(self, section, orientation, role):
        result = QtCore.QVariant()
        if (orientation == QtCore.Qt.Horizontal and
                (role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole)):
            result = ['Parameter', 'Type', 'Value'][section]

        return result

    # Called after user edits parameter value to keep model in sync
    def setData(self, mod_index, value, role=QtCore.Qt.EditRole):
        row = mod_index.row()
        param_dict = self._element.param_list[row]
        success = True

        # This is a little tricky - if user enters valid value assign it to
        # 'value' in the param dictionary.  If user enters nothing assign
        # 'default' to 'value' if 'default' exists, otherwise delete 'value'
        # from param dictionary.  Talk to Cole if you don't like this.
        if role == QtCore.Qt.EditRole:
            if value.strip():
                success, cvt_value = self._convert_value(
                    param_dict['type'], value)
                if success:
                    param_dict['value'] = cvt_value
            else:
                if 'default' in param_dict.keys():
                    param_dict['value'] = param_dict['default']
                else:
                    param_dict.pop('value', None)

        return success

    #
    # Helper Methods
    #

    def _convert_value(self, type_str, value_str):
        success, cvt_value = True, None
        try:
            if value_str == 'None':
                cvt_value = None

            elif type_str in ['str', 'file', 'folder', 'file_or_folder']:
                cvt_value = value_str

            elif type_str == 'int':
                if value_str == 'inf':
                    cvt_value = float(value_str)
                else:
                    cvt_value = int(value_str)

            elif type_str == 'float':
                cvt_value = float(value_str)

            elif type_str == 'int_list':
                cvt_value = self._str_list_to_int_list(value_str)

            elif type_str == 'int_list_list':
                # Strip outer sq brackets: '[[1,2],[3,4]]' -> '[1,2],[3,4]'
                value_str = re.sub(r'^\[|\]$', '', value_str)
                # Disambiguate list separator: '[1,2],[3,4]' -> [1,2]:[3,4]'
                value_str = re.sub(r'\] *, *\[', ':', value_str)
                # Split into int_list strings and convert into int_lists
                cvt_value = list(map(
                    self._str_list_to_int_list, value_str.split(':')))

            elif type_str == 'bool':
                if value_str.lower() in ['true', 'yes']:
                    cvt_value = True
                elif value_str.lower() in ['false', 'no']:
                    cvt_value = False
                else:
                    raise TypeError(f'{value_str} is not a valid bool type')

            elif type_str == 'np.dtype':
                # Conversion test will trigger exception if not well-formed
                np.dtype(value_str)
                # Save np.dtype as string for consumption by SpikeInterface
                cvt_value = value_str

            else:
                raise TypeError(f'{type_str} is not a Spikely supported type')

        except (TypeError, ValueError) as err:
            QtWidgets.QMessageBox.warning(
                config.get_main_window(), 'Type Conversion Error', repr(err))
            success = False

        return success, cvt_value

    # E.g., '[1,2,3]' -> [1,2,3]
    def _str_list_to_int_list(self, str_list):
        # Set up split by stripping square brackets
        str_list = re.sub(r'^\[|\]$', '', str_list)
        # Convert list of string ints into actual ints
        return list(map(int, str_list.split(',')))



from . import element_policy as sp_elp
from . import recording_extractor as sp_ree
from . import sorter as sp_sor
from . import preprocessor as sp_pre
from . import curator as sp_cur
from . import sorting_exporter as sp_soe


class StdElementPolicy(sp_elp.ElementPolicy):

    def __init__(self):
        required_cls_list = [
            sp_ree.RecordingExtractor, sp_sor.Sorter]

        cls_order_dict = {
            sp_ree.RecordingExtractor: 0, sp_pre.Preprocessor: 1,
            sp_sor.Sorter: 2, sp_cur.Curator: 3, sp_soe.SortingExporter: 4}

        cls_display_name_dict = {
            sp_ree.RecordingExtractor: 'Recording Extractor',
            sp_pre.Preprocessor: 'Preprocessor',
            sp_sor.Sorter: 'Sorter', sp_cur.Curator: 'Curator',
            sp_soe.SortingExporter: 'Sorting Exporter'}

        super().__init__(required_cls_list, cls_order_dict,
                         cls_display_name_dict)

    def is_cls_available(self, cls):
        return cls in [
            sp_ree.RecordingExtractor, sp_pre.Preprocessor, sp_sor.Sorter,
            sp_cur.Curator, sp_soe.SortingExporter]

    def is_cls_singleton(self, cls):
        return cls in [
            sp_ree.RecordingExtractor, sp_sor.Sorter, sp_soe.SortingExporter]



# Make version id available at a package level
from .elements.spike_element import SpikeElement            # noqa: F401
from .elements.std_element_policy import StdElementPolicy   # noqa: F401
from .pipeline_model import PipelineModel                   # noqa: F401
from .pipeline_view import PipelineView                     # noqa: F401
from .parameter_view import ParameterView                   # noqa: F401
from .parameter_model import ParameterModel                 # noqa: F401
from .operation_view import OperationView                   # noqa: F401
from .version import __version__                            # noqa: F401



from . import spike_element as sp_spe
import spikesorters as ss

from PyQt5 import QtGui
from PyQt5 import QtWidgets
import pkg_resources

import copy


class Sorter(sp_spe.SpikeElement):
    @staticmethod
    def get_installed_spif_cls_list():
        return ss.installed_sorter_list

    @staticmethod
    def get_display_name_from_spif_class(spif_class):
        return spif_class.sorter_name

    def __init__(self, spif_class):
        super().__init__(spif_class)

        self._display_name = self.get_display_name_from_spif_class(spif_class)

        if QtWidgets.QApplication.instance():
            self._display_icon = QtGui.QIcon(
                pkg_resources.resource_filename(
                    'spikely.resources', 'sorter.png'))
        else:
            self._display_icon = None

        self._param_list = copy.deepcopy(spif_class.sorter_gui_params)

    @property
    def display_name(self):
        return self._display_name

    @property
    def display_icon(self):
        return self._display_icon

    def run(self, payload, next_elem):

        spif_param_list = {param['name']: param['value'] for param
                           in self._param_list if param.get('base_param')}
        spif_param_list['recording'] = payload
        sorter = self._spif_class(**spif_param_list)

        sub_param_list = {param['name']: param['value'] for param
                          in self._param_list if not param.get('base_param')}
        sorter.set_params(**sub_param_list)

        sorter.run()

        output_folder_string = sub_param_list.get('output_folder')
        if output_folder_string is None:
            output_folder_string = 'tmp_' + sorter.sorter_name

        return sorter.get_result_list(), output_folder_string, payload



import spikeextractors as se
from .phy_exporter import PhyExporter

exporters_list = []
exporters_list.extend(se.extractorlist.writable_sorting_extractor_list)
exporters_list.append(PhyExporter)

se.sorting_exporter_dict

sorting_exporter_dict = se.sorting_exporter_dict
sorting_exporter_dict[PhyExporter.exporter_name] = PhyExporter



from setuptools import setup, find_packages

version = {}
with open('spikely/version.py') as fp:
    exec(fp.read(), version)

with open('README.md', 'r') as fp:
    long_description = fp.read()

setup(
    name='spikely',
    version=version['__version__'],
    description='Spike sorting made simple',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Roger Hurwitz',
    author_email='rogerhurwitz@gmail.com',
    url='https://github.com/SpikeInterface/spikely',
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
    install_requires=[
        'pyqt5>=5.13.1',
        'spikeextractors>=0.7.0',
        'spikesorters>=0.2.0'
    ],
    entry_points={
        'console_scripts': [
            'spikely=spikely.spikely_main:launch_spikely'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
)



import json

import pkg_resources
from PyQt5 import QtCore, QtWidgets

from spikely import SpikeElement, StdElementPolicy, config


class PipelineModel(QtCore.QAbstractListModel):

    def __init__(self, parameter_model):
        super().__init__()

        self._element_list = []
        self._element_policy = StdElementPolicy()
        self._parameter_model = parameter_model
        self._share_output = False

    # TODO: Put back in when ready to support terminal output
    @property
    def share_output(self):
        return self._share_output

    @share_output.setter
    def share_output(self, state):
        self._share_output = state

    def _elem_cls_count(self, target_cls):
        elem_cls_list = [type(elem) for elem in self._element_list]
        return elem_cls_list.count(target_cls)

    def rowCount(self, parent=None):
        # Overrides base class: provides count of elements in pipeline
        return len(self._element_list)

    def data(self, mod_index, role=QtCore.Qt.DisplayRole):
        # Overrides base class: returns data for element in pipeline for role
        if mod_index.isValid() and mod_index.row() < len(self._element_list):
            element = self._element_list[mod_index.row()]
            data_dict = {
                QtCore.Qt.DisplayRole:      element.display_name,
                QtCore.Qt.EditRole:         element.display_name,
                QtCore.Qt.DecorationRole:   element.display_icon,
                config.ELEMENT_ROLE:    element}
            return data_dict.get(role)

    def run(self):
        # Called in response to user pressing Run button in UI
        missing_param_count = self._missing_param_count()
        if missing_param_count:
            QtWidgets.QMessageBox.warning(
                config.get_main_window(), 'Run Failure',
                f'Missing mandatory element parameters.  Missing parameter '
                f'count: {missing_param_count}')
            return

        for cls in self._element_policy.required_cls_list:
            if not self._elem_cls_count(cls):
                QtWidgets.QMessageBox.warning(
                    config.get_main_window(), 'Run Failure',
                    f'Missing required element: {cls.__name__}')
                return

        config.get_main_window().statusBar().showMessage(
            'Running pipeline', config.STATUS_MSG_TIMEOUT)

        elem_jdict_list = [config.cvt_elem_to_dict(element)
                           for element in self._element_list]

        elem_list_str = json.dumps(elem_jdict_list)

        # TODO: Add plumbing for shared output support
        if self.share_output:
            run_path = pkg_resources.resource_filename(
                'spikely.pipeman', 'piperun.py')
        else:
            run_path = pkg_resources.resource_filename(
                'spikely.pipeman', 'pipeman.py')

        run_process = QtCore.QProcess(self)
        success = run_process.startDetached(
            'python', [f'{run_path}', elem_list_str])
        if not success:
            QtWidgets.QMessageBox.warning(
                config.get_main_window(), 'Failed to Start Python Process',
                f'Command line: python {run_path}, elem_list_str')

    def clear(self):
        self.beginResetModel()
        self._element_list.clear()
        self.endResetModel()
        # Synchronize parameter model and view
        self._parameter_model.element = None

    def add_element(self, new_elem: SpikeElement) -> None:
        # Inserts new element into  pipeline in proper order

        new_elem_cls_count = self._elem_cls_count(new_elem.__class__)
        new_elem_is_singleton = self._element_policy.\
            is_cls_singleton(new_elem.__class__)
        if new_elem_is_singleton and new_elem_cls_count:
            config.get_main_window().statusBar().showMessage(
                'Only one element of this type allowed in pipeline',
                config.STATUS_MSG_TIMEOUT)
            return

        target_positions = self._element_policy.cls_order_dict
        new_elem_target_pos = target_positions[new_elem.__class__]
        new_elem_insert_pos = 0
        for pipe_elem in self._element_list:
            pipe_elem_target_pos = target_positions[pipe_elem.__class__]
            if new_elem_target_pos >= pipe_elem_target_pos:
                new_elem_insert_pos += 1
            else:
                break

        self.beginInsertRows(QtCore.QModelIndex(), new_elem_insert_pos,
                             new_elem_insert_pos)
        self._element_list.insert(new_elem_insert_pos, new_elem)
        self.endInsertRows()

    # TODO: Clean this up in line w/ add_element method
    def move_up(self, elem: SpikeElement) -> None:
        rank = self._element_policy.cls_order_dict
        row = self._element_list.index(elem)

        if row > 0 and rank[type(elem)] == \
                rank[type(self._element_list[row - 1])]:
            self.beginMoveRows(QtCore.QModelIndex(), row,
                row, QtCore.QModelIndex(), row - 1)  # noqa: E128
            self._swap(self._element_list, row, row - 1)
            self.endMoveRows()
        else:
            config.get_main_window().statusBar().showMessage(
                "Cannot move element any higher", config.STATUS_MSG_TIMEOUT)

    # TODO: Clean this up in line w/ add_element method
    def move_down(self, elem: SpikeElement) -> None:
        rank = self._element_policy.cls_order_dict
        row = self._element_list.index(elem)

        if row < len(self._element_list) - 1 and \
                rank[type(elem)] == rank[type(self._element_list[row + 1])]:
            self.beginMoveRows(QtCore.QModelIndex(), row + 1,
                row + 1, QtCore.QModelIndex(), row)   # noqa: E128
            self._swap(self._element_list, row, row + 1)
            self.endMoveRows()
        else:
            config.get_main_window().statusBar().showMessage(
                "Cannot move element any lower", config.STATUS_MSG_TIMEOUT)

    def delete(self, element: SpikeElement) -> None:
        index = self._element_list.index(element)
        self.beginRemoveRows(QtCore.QModelIndex(), index, index)
        self._element_list.pop(index)
        self.endRemoveRows()

    # Helper methods

    def _swap(self, list, pos1, pos2):
        list[pos1], list[pos2] = list[pos2], list[pos1]

    def _missing_param_count(self) -> int:
        missing_param_list = [param for elem in self._element_list for param in
                              elem.param_list if 'value' not in param.keys()]
        return len(missing_param_list)



"""Constructs File menu/actions, and executes user selected actions

During application initialization in spikely_main a menu bar is created as part
of the main window.  create_file_menu is called as part of that process to
populate the menu bar with a drop down menu of file related actions.

The core functionality within this module is support for saving and loading the
elements in the pipeline as JSON files. In collaboration with config.py, which
has methods to convert individual elements to/from serializable dictionary
objects, the _perform load/save methods in this module operate on the pipeline
as a whole.

"""

import json

from PyQt5 import QtWidgets

from spikely import PipelineModel, config

# Enables access to element list for both input and output
_pipeline_model = None


# Menu and Menu Action construction methods
def create_file_menu(main_window: QtWidgets.QMainWindow,
                     pipeline_model: PipelineModel) -> QtWidgets.QMenu:
    global _pipeline_model
    _pipeline_model = pipeline_model

    file_menu = QtWidgets.QMenu('&File', main_window)

    file_menu.addAction(_action(
        'Load Pipeline', 'Load pipeline from JSON file', _perform_load_action))
    file_menu.addAction(_action(
        'Save Pipeline', 'Save pipeline to JSON file', _perform_save_action))
    file_menu.addSeparator()
    file_menu.addAction(_action(
        'Share Output', 'Use terminal for all pipeline output',
        _toggle_share_state, checkable=True, checked=False))
    file_menu.addSeparator()
    file_menu.addAction(_action(
        'Exit', 'Terminate the application.',
        QtWidgets.QApplication.closeAllWindows))

    return file_menu


def _action(name, tip, slot, shortcut=None, checkable=False, checked=None):
    action = QtWidgets.QAction(name, config.get_main_window(),
                               checkable=checkable)
    action.setStatusTip(tip)
    action.triggered.connect(slot)
    if shortcut is not None:
        action.setShortcut(shortcut)
    if checkable and checked is not None:
        action.setChecked(checked)

    return action


# Menu Action execution methods

def _toggle_share_state(checked):
    _pipeline_model.share_output = checked


def _perform_load_action() -> None:
    """Loads current pipeline with elements from a previously saved JSON file

    Launches a file dialog box that allows the user to select a previously
    saved JSON file, attempts to decode it, and if successful adds the elements
    to the current pipeline replacing any elements extant in the pipeline.

    config.cvt_dict_to_elem() does most of the hard work, and throws exceptions
    if the element is no longer installed, or is no longer compatible with the
    version saved previously.

    """
    global _pipeline_model

    options = QtWidgets.QFileDialog.Options()
    options |= QtWidgets.QFileDialog.DontUseNativeDialog
    file_name, _filter = QtWidgets.QFileDialog.getOpenFileName(
            config.get_main_window(), caption='Open File',
            filter='JSON (*.json)', options=options)

    if file_name:
        _pipeline_model.clear()
        try:
            with open(file_name, 'r') as json_file:
                elem_dict_list = json.load(json_file)

            for elem_dict in elem_dict_list:
                elem = config.cvt_dict_to_elem(elem_dict)
                _pipeline_model.add_element(elem)

        except (json.decoder.JSONDecodeError, ValueError) as e:
            QtWidgets.QMessageBox.warning(
                config.get_main_window(), 'JSON File Load Failure',
                f'Failed to load {file_name}: {str(e)}')
            _pipeline_model.clear()

        except Exception as e:
            QtWidgets.QMessageBox.warning(
                config.get_main_window(), 'JSON File Load Failure',
                f'Unspecified exception: {str(e)}')
            _pipeline_model.clear()


def _perform_save_action() -> None:
    """Saves current pipeline of elements to a user specified JSON file

    Launches a file dialog box that allows the user to specifiy a JSON file,
    attempts to encode the element pipeline, and if successful writes out the
    decoded element pipelin in JSON format to file.

    config.cvt_elem_to_dict() does most of the hard work, by extracting class
    and parameter data from the element that allows the element to be JSON
    encoded and reinstantiated later when the filed is read back in by
    _perform_load_action().

    """
    global _pipeline_model

    # TODO: _element_list is supposed to be private - use data() instead?
    element_list = _pipeline_model._element_list

    if element_list:
        options = QtWidgets.QFileDialog.Options()
        options |= QtWidgets.QFileDialog.DontUseNativeDialog
        file_name, _filter = QtWidgets.QFileDialog.getSaveFileName(
            config.get_main_window(), caption='Save File',
            filter='JSON (*.json)', options=options)

        if file_name:
            if not file_name.lower().endswith('.json'):
                file_name = file_name + '.json'
            elem_dict_list = [config.cvt_elem_to_dict(element)
                              for element in element_list]

            with open(file_name, 'w') as json_file:
                json.dump(elem_dict_list, json_file)



from PyQt5 import QtWidgets

from spikely import config

# TODO: Implement instance (versus class) version of QFileDialog


# Menu and Menu Action construction methods
def create_tool_bar(main_win):
    tool_bar = QtWidgets.QToolBar(main_win)
    tool_bar.setMovable(False)
    tool_bar.setFloatable(False)

    folder_act = QtWidgets.QAction(QtWidgets.QFileIconProvider().icon(
        QtWidgets.QFileIconProvider.Folder), 'Select Folder', main_win)
    folder_act.setStatusTip('Choose folder and copy path into clipboard '
        'to enable pasting path into an element parameter field')  # noqa: E128
    folder_act.triggered.connect(_perform_folder_action)
    tool_bar.addAction(folder_act)

    file_act = QtWidgets.QAction(QtWidgets.QFileIconProvider().icon(
        QtWidgets.QFileIconProvider.File), 'Select File', main_win)
    file_act.setStatusTip('Choose file and copy path into clipboard '
        'to enable pasting path into an element parameter field')  # noqa: E128
    file_act.triggered.connect(_perform_file_action)
    tool_bar.addAction(file_act)

    return tool_bar


def _perform_file_action() -> None:

    options = QtWidgets.QFileDialog.Options()
    options |= QtWidgets.QFileDialog.DontUseNativeDialog
    file_name, _filter = QtWidgets.QFileDialog.getOpenFileName(
            config.get_main_window(), caption='Copy File Name to Clipboard',
            options=options)

    if file_name:
        QtWidgets.QApplication.clipboard().setText(file_name)


def _perform_folder_action() -> None:

    options = QtWidgets.QFileDialog.Options()
    options |= QtWidgets.QFileDialog.DontUseNativeDialog
    options |= QtWidgets.QFileDialog.ShowDirsOnly
    options |= QtWidgets.QFileDialog.DontResolveSymlinks
    folder_name = QtWidgets.QFileDialog.getExistingDirectory(
            config.get_main_window(), caption='Copy Folder Name to Clipboard',
            options=options)

    if folder_name:
        QtWidgets.QApplication.clipboard().setText(folder_name)


'''
    # tool_menu = menu_bar.addMenu(qw.QMenu('Tools', self))
    # dir_action = qw.QAction('Pick Directory', self)
    # dir_action.setShortcut('Ctrl+D')
    # dir_action.setStatusTip('Copy directory path to clipboard')
    # dir_action.triggered.connect(self.do_dir_action)
    # tool_menu.addAction(dir_action)

   # def do_dir_action(self):
    #     dlg = qw.QFileDialog(self)
    #     dlg.setFileMode(dlg.Directory)
    #     dlg.setViewMode(dlg.List)
    #     dlg.setDirectory('.')
    #     dlg.setOption(dlg.DontUseNativeDialog, True)
    #     # dlg.setOption(dlg.ShowDirsOnly, True)
    #     dlg.setOption(dlg.ReadOnly, True)
    #     dlg.setOption(dlg.HideNameFilterDetails, True)

    #     if (dlg.exec_()):
    #         file_names = dlg.selectedFiles()
    #         cb = qw.QApplication.clipboard()
    #         cb.setText(file_names[0])
'''



from abc import ABC, abstractmethod


class SpikeElement(ABC):

    # Abstract methods

    @staticmethod
    @abstractmethod
    def get_installed_spif_cls_list():
        pass

    @staticmethod
    @abstractmethod
    def get_display_name_from_spif_class(spif_class):
        pass

    @abstractmethod
    def run(self, payload, next_elem):
        pass

    @property
    @abstractmethod
    def display_name(self):
        pass

    @property
    @abstractmethod
    def display_icon(self):
        pass

    # Concrete base class methods

    def __init__(self, spif_class):
        self._spif_class = spif_class
        self._param_list = None

    @property
    def spif_class(self):
        return self._spif_class

    @property
    def param_list(self):
        return self._param_list

    @param_list.setter
    def param_list(self, param_list):
        self._param_list = param_list



from PyQt5 import QtWidgets

from spikely.config import PARAM_COL, TYPE_COL, VALUE_COL


class ParameterView(QtWidgets.QGroupBox):

    def __init__(self, pipeline_model, parameter_model):
        super().__init__("Configure Parameters")
        self._pipeline_model = pipeline_model
        self._parameter_model = parameter_model

        self._init_ui()

    def _init_ui(self):
        self.setLayout(QtWidgets.QHBoxLayout())

        cfg_table = QtWidgets.QTableView(self)

        # Magic happens here: links element model to view
        cfg_table.setModel(self._parameter_model)

        cfg_table.verticalHeader().hide()
        cfg_table.setColumnWidth(PARAM_COL, 200)
        cfg_table.setColumnWidth(TYPE_COL, 100)
        cfg_table.setColumnWidth(VALUE_COL, 200)
        cfg_table.horizontalHeader().setStretchLastSection(True)

        self.layout().addWidget(cfg_table)
